// DTDState : Contains the active state of Direct3D.
//

#include "stdafx.h"
#include "DTDState.h"

// Global Direct Three Dimension status pointer.
DTDState* pDTDState;

// Texture Creation (TC) function pointers.
VTexture* (*pTC_P_8)();			// Creates a texture capable of displaying a palettized 8bit texture.
VTexture* (*pTC_AP_88)();		// Creates a texture capable of displaying an alpha interlieved palettized 8bit texture.
VTexture* (*pTC_ARGB_4444)();	// Creates a texture capable of displaying a 16bit ARGB (4bit alpha and 4bpcc) texture.
VTexture* (*pTC_ARGB_1555)();	// Creates a texture capable of displaying a 16bit ARGB (1bit alpha and 5bpcc) texture.

//-----------------------------------
// Genuine-texture creation functions.

// Our null texture function.
VTexture* TC_Null()
{
	return new VTexture;
}

// Create a genuine palettized 8bit texture.
VTexture* TC_P_8()
{
	return new DTDTexture_P_8;
}

// Create a genuine alpha interlieved and palettized 8bit texture.
VTexture* TC_AP_88()
{
	return new DTDTexture_AP_88;
}

// Create a genuine ARGB_4444 texture.
VTexture* TC_ARGB_4444()
{
	return new DTDTexture_ARGB_4444;
}

//-----------------------------------
// Emulated-textures creation functions.
VTexture* TC_RGB_565_EmulateP8()
{
	return new DTDTexture_RGB_565_EmulateP8;
}

VTexture* TC_ARGB_4444_EmulateAP88()
{
	return new DTDTexture_ARGB_4444_EmulateAP88;
}

//-------------------------------------------------------------------
DTDState::DTDState()
{
	bSceneReady = false;
	eColorMode = kIterated;
	eAlphaMode = kIterated;
	dwConstantColor = 0xFFFFFFFF;
	dwConstantColorNoAlpha = dwConstantColor & 0x00FFFFFF;
	dwConstantAlpha = 0xFF000000;
	fZBias = 0;

	// Initialize the texture buffer on the next call to CreateAt();
	pTextureBuffer = 0;

	// Track and trash texture management.
	pFirstTexture = 0;
	pCurrentTexture = 0;

	// Default to having the palette flagged as clean.
	bPalDirty = false;

	// Chroma keying mode.
	dwChromaKeyColor = 0;

	bUseTextures = false;
	bUseBlending = false;
}

DTDState::~DTDState()
{
	Toast();
}

DWORD CountBits( DWORD value )
{
	int bits;
	for(bits=0; value; value>>=1)
	{
		if( value & 1)
			bits++;
	}
	return bits;
}


HRESULT CALLBACK SpewPixelFormatsCB( LPDDPIXELFORMAT lpDDPixFmt, LPVOID lpContext )
{
	FILE *pFile = (FILE*)lpContext;

	// Prepare a clean looking texture mark seperator.
	fprintf( pFile, "------------------------------------------------------------------\n"
					"Texture Pixel Format Description\n"
					"---------------------------------\n");

	// Spit out the texture flags.
	fprintf( pFile, "Flags (%u): ", lpDDPixFmt->dwFlags);
	if( lpDDPixFmt->dwFlags & DDPF_ALPHA )
	{
		fprintf( pFile, " DDPF_ALPHA" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_ALPHAPIXELS )
	{
		fprintf( pFile, " DDPF_ALPHAPIXELS" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_ALPHAPREMULT )
	{
		fprintf( pFile, " DDPF_ALPHAPREMULT" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_BUMPLUMINANCE )
	{
		fprintf( pFile, " DDPF_BUMPLUMINANCE" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_BUMPDUDV )
	{
		fprintf( pFile, " DDPF_BUMPDUDV" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_COMPRESSED )
	{
		fprintf( pFile, " DDPF_COMPRESSED" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_FOURCC )
	{
		fprintf( pFile, " DDPF_FOURCC" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_LUMINANCE )
	{
		fprintf( pFile, " DDPF_LUMINANCE" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_PALETTEINDEXED1 )
	{
		fprintf( pFile, " DDPF_PALETTEINDEXED1" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_PALETTEINDEXED2 )
	{
		fprintf( pFile, " DDPF_PALETTEINDEXED2" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_PALETTEINDEXED4 )
	{
		fprintf( pFile, " DDPF_PALETTEINDEXED4" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_PALETTEINDEXED8 )
	{
		fprintf( pFile, " DDPF_PALETTEINDEXED8" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_PALETTEINDEXEDTO8 )
	{
		fprintf( pFile, " DDPF_PALETTEINDEXEDTO8" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_RGB )
	{
		fprintf( pFile, " DDPF_RGB" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_RGBTOYUV )
	{
		fprintf( pFile, " DDPF_RGBTOYUV" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_STENCILBUFFER )
	{
		fprintf( pFile, " DDPF_STENCILBUFFER" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_YUV )
	{
		fprintf( pFile, " DDPF_YUV" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_ZBUFFER )
	{
		fprintf( pFile, " DDPF_STENCILBUFFER" );
	}

	if( lpDDPixFmt->dwFlags & DDPF_ZPIXELS )
	{
		fprintf( pFile, " DDPF_ZPIXELS" );
	}

	// Spill the beens on the FourCC info.
	fprintf( pFile, "\n--\nFourCC: %u\n", lpDDPixFmt->dwFourCC );

	// Crack the bit buckets for RGB/YUV/ZBuffer info.
	fprintf( pFile, "--\nBitCount: %d\n", lpDDPixFmt->dwRGBBitCount );

	fprintf( pFile, " R Mask (%2d): 0x%08X\n", CountBits(lpDDPixFmt->dwRBitMask), lpDDPixFmt->dwRBitMask );
	fprintf( pFile, " G Mask (%2d): 0x%08X\n", CountBits(lpDDPixFmt->dwGBitMask), lpDDPixFmt->dwGBitMask );
	fprintf( pFile, " B Mask (%2d): 0x%08X\n", CountBits(lpDDPixFmt->dwBBitMask), lpDDPixFmt->dwBBitMask );
	fprintf( pFile, " A Mask (%2d): 0x%08X\n", CountBits(lpDDPixFmt->dwRGBAlphaBitMask), lpDDPixFmt->dwRGBAlphaBitMask );
	fprintf( pFile, "\n\n" );


	// Figure out if we have an exact match to one of the supported native texture modes.
	if( lpDDPixFmt->dwRGBBitCount == 8 )
	{ // An 8bit texture mode.
		if( lpDDPixFmt->dwFlags == (DDPF_PALETTEINDEXED8 | DDPF_RGB) )
		{ // Yeah!  We have a palettized 8bit texture mode here!
			pTC_P_8 = TC_P_8; // Native 8bit palettized texutre mode.
		}
	}

	// Keep going, right to the end.
	return D3DENUMRET_OK;
};

// Prepare the texturing configuration for whatever the video hardware is capable of using.
void DTDState::PrepareTextureModes()
{
	// Setup the texture creation pointers to the most compatible methods available.
	pTC_P_8 = TC_RGB_565_EmulateP8;			// Oh, the illusive palettized texture mode.
	pTC_AP_88 = TC_ARGB_4444_EmulateAP88;	// And the currently-not-implemented-by-anyone-yet alpha interleave mode.
										 	// (Heck even 3Dfx didn't implement any of their own supported texture modes in D3D)
	pTC_ARGB_4444 = TC_ARGB_4444;	// This mode seems to be supported by just about everyone.
	pTC_ARGB_1555 = TC_Null;		// This mode has yet to be supported by DirectGlide.



	FILE *pTexOutFile;
	pTexOutFile = fopen( "TextureModes.txt", "wt" );
	Framework.GetD3DDevice()->EnumTextureFormats( SpewPixelFormatsCB, pTexOutFile );
	fclose( pTexOutFile );
}

// Destroy all internal objects.
void DTDState::Toast()
{
	// Nuke the texture linked list.
	VTexture *pTex = pFirstTexture;
	for( ; pTex; pTex=pTex->m_pNext )
	{
		pTex->Toast();
		delete pTex;
	}

	pFirstTexture = 0;
}


//-------------------------------------------------------------------
// Scene control functions, useful for making sure
// the BeginScene and EndScene pairs are always
// correctly synched.
//-----------------------------------
// Makes sure the Direct3D drawing features are in BeginScene state.
void DTDState::BeginScene()
{
	pGlideState->FlushStates();

	if( !bSceneReady )
	{
		// Start our scene.
		pDTDState->Framework.GetD3DDevice()->BeginScene();
		bSceneReady = true;
	}
}

//-----------------------------------
// Makes sure the Direct3D drawing features are in EndScene state.
void DTDState::EndScene()
{
	if( bSceneReady )
	{
		// Start our scene.
		pDTDState->Framework.GetD3DDevice()->EndScene();
		bSceneReady = false;
	}
}


//-------------------------------------------------------------------
// Texture management functions.
//-------------------------------------------------------------------
// Texture Class.

// Here, I grabbed a copy of Let's code to translate some useful info.
// Convert the LOD info to an integer.
int ConvLodToInt(GrLOD_t lod)
{
	switch (lod) {
	case GR_LOD_1  : return 1;
	case GR_LOD_2  : return 2;
	case GR_LOD_4  : return 4;
	case GR_LOD_8  : return 8;
	case GR_LOD_16 : return 16;
	case GR_LOD_32 : return 32;
	case GR_LOD_64 : return 64;
	case GR_LOD_128: return 128;
	case GR_LOD_256: return 256;
	}
	return 0;
}

// Here, I grabbed a copy of Let's code to translate some useful info.
// Convert the integer to an LOD state value.
GrLOD_t ConvIntToLod(int nLod)
{
	switch (nLod) {
	case 1   : return GR_LOD_1;
	case 2   : return GR_LOD_2;
	case 4   : return GR_LOD_4;
	case 8   : return GR_LOD_8;
	case 16  : return GR_LOD_16;
	case 32  : return GR_LOD_32;
	case 64  : return GR_LOD_64;
	case 128 : return GR_LOD_128;
	case 256 : return GR_LOD_256;
	}
	return 0;
}

// Here, I grabbed a copy of Let's code to translate some useful info.
// Convert the aspect ratio into Width and Height info.
void ConvAspectToInt(int &nAspectW, int &nAspectH, GrAspectRatio_t aspect_ratio)
{
	switch (aspect_ratio) {
	case GR_ASPECT_8x1: nAspectW = 8; nAspectH = 1; return;
	case GR_ASPECT_4x1: nAspectW = 4; nAspectH = 1; return;
	case GR_ASPECT_2x1: nAspectW = 2; nAspectH = 1; return;
	case GR_ASPECT_1x1: nAspectW = 1; nAspectH = 1; return;
	case GR_ASPECT_1x2: nAspectW = 1; nAspectH = 2; return;
	case GR_ASPECT_1x4: nAspectW = 1; nAspectH = 4; return;
	case GR_ASPECT_1x8: nAspectW = 1; nAspectH = 8; return;
	}
}

// Here, I grabbed a copy of Let's code to translate some useful info.
// Convert a Width and Height into an aspect ratio state value.
void ConvIntToAspect( GrAspectRatio_t &aspect_ratio, int nAspectW, int nAspectH)
{
	if (nAspectW == 1) {
		if (nAspectH == 1) aspect_ratio = GR_ASPECT_1x1; return;
		if (nAspectH == 2) aspect_ratio = GR_ASPECT_1x2; return;
		if (nAspectH == 4) aspect_ratio = GR_ASPECT_1x4; return;
		if (nAspectH == 8) aspect_ratio = GR_ASPECT_1x8; return;
	} else {
		if (nAspectW == 2) aspect_ratio = GR_ASPECT_2x1; return;
		if (nAspectW == 4) aspect_ratio = GR_ASPECT_4x1; return;
		if (nAspectW == 8) aspect_ratio = GR_ASPECT_8x1; return;
	}
}

void DTDState::SetTextureRange( VTexture *pTexture, DWORD iStart)
{
	pTextureBuffer[iStart/8] = pTexture;
}

//-------------------------------------------------------------------
// Texture management methods.
VTexture* DTDState::CreateAt( DWORD iStart, DWORD iEnd, GrTexInfo *pInfo )
{
	if(pTextureBuffer == 0)
	{ // Make sure we have a texture buffer to work with.
		pTextureBuffer = new VTexture*[pGlideState->nMaxTextureMemory/8];
		for( DWORD iTB=0; iTB<(pGlideState->nMaxTextureMemory/8); iTB++)
			pTextureBuffer[iTB] = 0;
	}

	// Make sure the primitive buffers are empty.
	pDTDState->Framework.GetD3DDevice()->SetRenderState( D3DRENDERSTATE_FLUSHBATCH, TRUE );

	// Trash the inclusive range of this new texture.
	TrashTextureRange( iStart, iEnd );

	// TBI - Implement the creation pointer methods, needed for each texture type.
	// We need something like the color conversion function pointers, located
	// in DTDState.h, but for creating the new texture instances.
	VTexture *pTex = 0;
	switch( pInfo->format )
	{ // I've only implemented the textures used by Tribes.
	case GR_TEXFMT_P_8:
		pTex = pTC_P_8();
		break;

	case GR_TEXFMT_ARGB_1555:
		pTex = pTC_ARGB_1555();
		break;

	case GR_TEXFMT_ARGB_4444:
		pTex = pTC_ARGB_4444();
		break;

	case GR_TEXFMT_AP_88:
		pTex = pTC_AP_88();
		break;

	default:
		pTex = new VTexture;
		break;
	}

	// Abort on texture creation failure.
	if( pTex == 0 )
	{
		DebugOut( "DTDState::CreateAt() - Failed to create texture of type: %d\n", pInfo->format);
		return 0;
	}

	if( pCurrentTexture == 0 )
	{	// Had problems with the CurrentTexture becoming invalid, so in those
		// cases, we just reassociate with this new texture.
		pCurrentTexture = pTex;
	}

	// Fill in all the necessary info.
	memcpy( &pTex->m_Info, pInfo, sizeof(GrTexInfo) );
	pTex->m_iBegin = iStart;
	pTex->m_iEnd = iEnd;

	// Fill in the given texture range in our texture buffer.
	SetTextureRange( pTex, iStart );

	//----------
	// Calculate the texture width and height at the highest mipmap level.
	// Large lod
	int nLargeLod;
	nLargeLod = ConvLodToInt(pInfo->largeLod);

	// Aspect ratio
	int nAspectW, nAspectH;
	ConvAspectToInt(nAspectW, nAspectH, pInfo->aspectRatio);

	// Width and Height
	// Aspects are swapped to produce the correct result.  Remember, we want the largest value relating to the largest aspect.
	pTex->nWidth = nLargeLod / nAspectH;
	pTex->nHeight = nLargeLod / nAspectW;

	// Create the texture surface info, with the information we already gave it.
	pTex->Create();

	if( pFirstTexture )
	{ // We have a list of textures, so insert our newest texture in the list.
		VTexture *pTmp = pFirstTexture, *pLast;

		// Find the first texture that is right before our start range.
		for( ; pTmp; pTmp=pTmp->m_pNext )
		{
			if( pTmp->m_iBegin > iStart )
				break; // Found an iStart which extended past the iStart we're after.

			// Record the last texture. (For end-of-chain cases)
			pLast = pTmp;
		}

		// Link the new texture into the appropriate texture position.
		if( pTmp == pFirstTexture )
		{	// The new texture sits in front of the first texture in the list, so replace its first position.
			// pFirstTexture=pTex 0<-[pTex]-> <-[pTmp]-> <-[...]->0
			pFirstTexture = pTex;	// Take the first texture position.
			pTex->m_pPrev = 0;		// Reset the new texture's prev pointer, cause we're at the top here!
			pTex->m_pNext = pTmp;	// /\ Link the new onto the next texture.
			pTmp->m_pPrev = pTex;	// \/ Link the next onto the new texture.
		} else if( pTmp )
		{	// We're sitting in the middle of the list somewhere, so insert the new texture.
			// [pLast]-> <-[pTex]-> <-[pTmp]-> <-[...]->0
			pLast->m_pNext = pTex;	// /\ Link the last onto the new texture.
			pTex->m_pPrev = pLast;	// \/ Link the new onto the last texture.
			pTex->m_pNext = pTmp;	// /\ Link the new onto the next texture.
			pTmp->m_pPrev = pTex;	// \/ Link the next onto the new texture.
		} else
		{	// Attach the texture to the end of the whole list.
			// [pLast]-> <-[pTex]->0
			pLast->m_pNext = pTex;	// /\ Link the last onto the new texture.
			pTex->m_pPrev = pLast;	// \/ Link the new onto the last texture.
			pTex->m_pNext = 0;		// Reset the new texture's next pointer, cause we're at the bottom here!
		}
	} else
	{	// Easy case, we're creating a new list of textures, so toss it in the First texture pointer.
		pFirstTexture = pTex;
		pTex->m_pPrev = 0;		// Reset the new texture's prev pointer, cause we're at the top here!
		pTex->m_pNext = 0;		// Reset the new texture's next pointer, cause we're at the bottom here!
	}

	return pTex;
}

// Locate a texture at a given address.
static int gTextureCounter = 0;
VTexture* DTDState::GetTextureAt( DWORD addressStart )
{
	if( pTextureBuffer )
	{
		return pTextureBuffer[addressStart/8];
	} else
		return 0;
}

// Remove textures in a given range.
void DTDState::TrashTextureRange( DWORD iStart, DWORD iEnd )
{
	VTexture *pTex = pFirstTexture;
	VTexture *pNextTmp = pFirstTexture;

	// Find the first texture that is within our start range.
	for( ; pTex; pTex=pTex->m_pNext )
	{
		if( pTex->m_iEnd >= iStart )
			break; // Found an iEnd which extended past the iStart we're after.
	}

	// Walk through each texture which overlaps iStart to iEnd, until we get past iEnd.
	for( ; pTex; pTex=pNextTmp )
	{
		// Check if we walked right past the iEnd range.
		if( pTex->m_iBegin > iEnd )
		{ // We're out of range, and we're done.
			return;
		}

		// Otherwise, we're overlapping this range in some way, so
		// grab the next texture, and nuke the current one.
		pNextTmp = pTex->m_pNext;

		//-----------------------------------
		// Detach the linked list element from the list.
		if( pTex->m_pPrev )
		{	// Forward the previous link to our next link.
			pTex->m_pPrev->m_pNext = pTex->m_pNext;
		} else
		{	// If we have no previous link, then we must be looking at the top list element.
			// Forward the top link to our next link.
			pFirstTexture = pTex->m_pNext;
		}

		if( pTex->m_pNext )
		{	// Rewind the next link to our prev link.
			pTex->m_pNext->m_pPrev = pTex->m_pPrev;
		}

		// Make sure our current texture is not in our trashing list.
		if( pTex == pCurrentTexture )
		{ // Reset the sucker to prevent invalid texture execution.
			pCurrentTexture = 0;
		}

		// For safety, wipe the given texture range in our texture buffer.
		SetTextureRange( 0, pTex->m_iBegin );

		// The list element is detatched, now we can safely toast the sucker.
		pTex->Toast();
		delete pTex;
	}
}
