//-------------------------------------------------------------------
//	File:		DTDTexture_RGB_565_EmulateP8.cpp
//	Created:	16/02/99 11:59PM
//	Author:		Aaron Hilton
//	Comments:	Implement the 8bit palettized texture format
//				through RGB_565 emulation.
//-------------------------------------------------------------------
#include "stdafx.h"
#include "DTDTexture_RGB_565_EmulateP8.h"

// TBI - Must implement proper MipMap support.

// Create the texture from a given (inclusive) address range.
// Refer to pGlideState->pTexMemory for the texture memory.
bool DTDTexture_RGB_565_EmulateP8::Create()
{
	// Relocate our m_Info's data pointer to the start of our texture frame buffer data.
	m_Info.data = &pGlideState->pTexMemory[m_iBegin];

	// Lay down the surface information. (Thanks go out to Let)
	DDSURFACEDESC2 surfDesc;
	ZeroMemory(&surfDesc, sizeof(DDSURFACEDESC2));
	surfDesc.dwSize = sizeof(DDSURFACEDESC2);	// Fill the structure's size info.
	surfDesc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT | DDSD_PIXELFORMAT;// | DDSD_MIPMAPCOUNT;	// Say we're going to use this info.
	surfDesc.ddsCaps.dwCaps = DDSCAPS_TEXTURE;
	surfDesc.ddsCaps.dwCaps2 = DDSCAPS2_TEXTUREMANAGE;	// Make this a self managed texture.

	// TBI - Prepare the real mipmap setting.
	surfDesc.dwMipMapCount = 1;

	// Prepare the texture format info.
	surfDesc.ddpfPixelFormat.dwSize = sizeof(DDPIXELFORMAT);
	surfDesc.dwWidth = nWidth;
	surfDesc.dwHeight = nHeight;

	// Configure the surface description for a palettized texture.
	DWORD nFormatWidth = nWidth;
	surfDesc.ddpfPixelFormat.dwFlags = DDPF_RGB;
	surfDesc.ddpfPixelFormat.dwRGBBitCount = 16;
	surfDesc.ddpfPixelFormat.dwRBitMask = 0x0000F800;
	surfDesc.ddpfPixelFormat.dwGBitMask = 0x000007E0;
	surfDesc.ddpfPixelFormat.dwBBitMask = 0x0000001F;
	surfDesc.ddpfPixelFormat.dwRGBAlphaBitMask = 0x00000000;

	// Yes, We do use a palette.
	bPal = true;

	// Attempt to create the texture's surface.
	if( pDTDState->Framework.GetDirectDraw()->CreateSurface(&surfDesc, &pSurface, 0 ) != DD_OK )
	{ // Failed to create the texture, so leave it at that.
		pSurface = 0;
		DebugOut( "DTDTexture_RGB_565_EmulateP8: Failed to create a texture surface.\n" );
		return false;
	}

	// Grab the texture interface.
	pSurface->QueryInterface(IID_IDirect3DTexture2, (LPVOID*)&pTexture);

	return true;
}

// Texture needs updating, so fill it in with the texture buffer.
void DTDTexture_RGB_565_EmulateP8::Update()
{
	// Wipe our palette to all black.
	ZeroMemory(m_PalEntries, sizeof(PALETTEENTRY)*256);

	// Force a dirty palette, which causes the update method below to be executed.
	pDTDState->bPalDirty = true;
}

// Announce that the palette needs updating.
// (Palette buffer is dirty, and we're about to draw!)
void DTDTexture_RGB_565_EmulateP8::UpdatePal()
{
	// Make sure we have a texture surface before updating.
	Assert( pSurface );

	// Compare the palettes to the internal palette.  If identical, then kick out of here, and ignore this update.
	bool bChanged = false;
	for( int iPal=0; iPal<256; iPal++)
	{
		if( *((DWORD*)(m_PalEntries+iPal)) != *((DWORD*)(pDTDState->pTexPalEntries+iPal)) )
		{ // Change detected, now replicate it into our own clean palette array, and mark a change.
			*((DWORD*)(m_PalEntries+iPal)) = *((DWORD*)(pDTDState->pTexPalEntries+iPal));
			bChanged = true;
		}
	}

	if( bChanged == false )
		return; // No changes found.


	// Make sure we're outside of a scene state.
	pDTDState->EndScene();

	// Handy variable, used by many other DTDTexture formats, helps with generalizing the code.
	DWORD nFormatWidth = nWidth;        // 8bpp textures have an equal format width, to their nWidth.

	// We have everything at this point, so let's transfer the relevent texture info, translating through our nifty palette.
	DDSURFACEDESC2 ddsd;
	ZeroMemory(&ddsd, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	pSurface->GetSurfaceDesc(&ddsd);
	if( pSurface->Lock(NULL, &ddsd, DDLOCK_WRITEONLY | DDLOCK_WAIT, NULL) != DD_OK )
	{ // Couldn't lock the texture's surface for the blitting, should be interesting. (Random data!)
		DebugOut( "DTDTexture_RGB_565_EmulateP8: Failed to lock the texture's surface for blitting.\n" );
		return;
	}

	// Blit the whole source texture data into the target surface.
	Assert( (DWORD)nHeight == ddsd.dwHeight );
	unsigned short *pDest;
	BYTE *pSrc;
	PALETTEENTRY *color;
	for( DWORD y=0; y<(DWORD)nHeight; y++ )
	{
		// Copy a scan line at a time.
		pDest = (unsigned short *)((BYTE*)ddsd.lpSurface + (y*ddsd.lPitch));
		pSrc = (BYTE*)m_Info.data + (y*nFormatWidth);
		for( DWORD x=0; x<(DWORD)nWidth; x++ )
		{
			color = &m_PalEntries[pSrc[x]];
			pDest[x] =	(unsigned short) \
						((((DWORD)color->peRed << 8 ) & 0xF800) +
						(((DWORD)color->peGreen << 3 ) & 0x07E0) +
						(DWORD)(color->peBlue >> 3 ));
		}
	}

	// Unlock the surface
	pSurface->Unlock(NULL);
}


//-------------------------------------------------------------------
//	History:
//	$Log:$
//-------------------------------------------------------------------
