// GlideState : Contains the active state of Glide.
//

#include "stdafx.h"
#include "GlideState.h"
#include "VertexProcessor.h"

GlideState* pGlideState;

GlideState::GlideState()
{
	pUniqueState = 0;
	fileUniqueStates = 0;

	// Setup and allocate the texture memory.
	nMaxTextureMemory = 4096*1024; // TBI - implement a smart texture buffer size based on the Direct3D HAL's capabilities.
	pTexMemory = new BYTE[nMaxTextureMemory];

	// Initialize the glide state structure to proper defaults.
	// Alpha blending.
	alpha_rgb_source	= GR_BLEND_ONE;
	alpha_rgb_dest		= GR_BLEND_ZERO;
	alpha_alpha_source	= GR_BLEND_ONE;
	alpha_alpha_dest	= GR_BLEND_ZERO;

	// Alpha combine processing.
	alpha_func		= GR_COMBINE_FUNCTION_SCALE_OTHER;
	alpha_factor	= GR_COMBINE_FACTOR_ONE;
	alpha_local		= GR_COMBINE_LOCAL_NONE;
	alpha_other		= GR_COMBINE_OTHER_CONSTANT;
	alpha_invert	= FXFALSE;

	// Hi bit iterated RGB lighting.  FX = Self Lit texture pixels.
	alpha_ITRGBLighting = FXFALSE;

	// Chromakey, in the alpha stage.
	alpha_test_func	= GR_CMP_ALWAYS;
	alpha_test_value	= 0;

	// Chromakey, in the pixelFX stage.
	chroma_mode	= GR_CHROMAKEY_DISABLE;
	chroma_value	= 0;

	// Clip Window.
	// TBI - these values should be configured in the SstWindowInit.
	clip_minx = 0;
	clip_miny = 0;
	clip_maxx = 640;
	clip_maxy = 480;

	// Color combine processing.
	color_func		= GR_COMBINE_FUNCTION_SCALE_OTHER;
	color_factor	= GR_COMBINE_FACTOR_ONE;
	color_local		= GR_COMBINE_LOCAL_NONE;
	color_other		= GR_COMBINE_OTHER_CONSTANT;
	color_invert	= FXFALSE;

	// Write masking enable/disable. (ColorMask, and DepthMask)
	mask_rgb	= FXTRUE;
	mask_alpha	= FXTRUE;

	// Constant color value.
	color_value = 0xFFFFFFFF;

	// Backface culling.
	cull_mode = GR_CULL_DISABLE;

	// Depth buffer control.
	depth_bias = 0;
	depth_func = GR_CMP_LESS;
	depth_mode = GR_DEPTHBUFFER_DISABLE;
	depth_mask	= FXFALSE;

	// Dither mode.
	dither_mode = GR_DITHER_4x4;

	// Fog control.
	fog_mode	= GR_FOG_DISABLE;
	fog_value	= 0;

	// Gamma correction value.
	gamma_value = 1.0;

	// Linear framebuffer control.
	lfb_alpha = 0xFF;
	lfb_depth = 0x00;

	// Rendering target buffer.
	render_buffer = GR_BUFFER_BACKBUFFER;

	// Texture clamping.
	tex_sClampMode = GR_TEXTURECLAMP_CLAMP;
	tex_tClampMode = GR_TEXTURECLAMP_CLAMP;
}

// Shutdown the class info.
GlideState::~GlideState()
{
	if( pTexMemory )
		delete pTexMemory;


	if( pUniqueState )
	{
		// We have unique states, so we must have a file open.
		fclose( fileUniqueStates );

		// Trash all the unique state classes.
		UniqueBlendState *pTmp;
		while( pUniqueState )
		{
			pTmp = pUniqueState->pNext;
			delete pUniqueState;
			pUniqueState = pTmp;
		}
	}
}

// Flush the states set by the DirtyFlags.
void GlideState::FlushStates()
{
	// Reject this call in most cases, because the render state doesn't change very frequently.
	if ( DirtyFlags == 0)
		return;
/*
	if( !pUniqueState )
	{
		// Open the file for the first time.
		fileUniqueStates = fopen( "UniqueStates.txt", "wt" );

		// Create our first unique blending state.
		pUniqueState = new UniqueBlendState;
		// Toss the new blending state to the file.
		pUniqueState->Print(fileUniqueStates);
	} else
	{
		UniqueBlendState *pNew = new UniqueBlendState;

		bool bIsUnique = true;

		// Verify that this is a unique blending state, before going further.
		for( UniqueBlendState *pIter=pUniqueState; pIter && bIsUnique ; pIter=pIter->pNext)
		{ // Walk through the unique states, verifying this one is not equal in any way.
			if( pNew->IsEqual( pIter ) )
			{ // Found a duplicate, so nuke the potential new one.
				bIsUnique = false;
				delete pNew;
			}
		}

		if( bIsUnique )
		{ // Attach the new state onto the end.
			pNew->pNext = pUniqueState;
			pUniqueState = pNew;

			// Toss the new blending state to the file.
			pNew->Print(fileUniqueStates);
		}
	}
*/

	if( DirtyFlags&(kAlpha | kColor | kMask) )
	{
		LPDIRECT3DDEVICE3 pDevice = pDTDState->Framework.GetD3DDevice();

		// Prepare the color processing stages.
		if( color_factor == GR_COMBINE_FACTOR_LOCAL )
		{ // We're dealing with a blending of two colors here.
			if( color_other == GR_COMBINE_OTHER_TEXTURE )
			{ // We're using blended texturing modes.
				pDTDState->bUseTextures = true;
				pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATE);

				if( color_local == GR_COMBINE_LOCAL_CONSTANT )
				{ // Use the constant factor mode.
					pDTDState->eColorMode = DTDState::EMode::kConstant;
				} else
				{ // Use the iterated diffuse mode.
					pDTDState->eColorMode = DTDState::EMode::kIterated;
				}
			} else
			{ // the color_other is always constant in the other case.
				pDTDState->bUseTextures = false;
				if( color_local == GR_COMBINE_LOCAL_CONSTANT )
				{ // Both constant colors, we can just use the FACTOR in D3D.
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
					pDTDState->eColorMode = DTDState::EMode::kConstant;
				} else
				{ // We're iterating against a constant value... tough.. real tough..
			  	// TBI - Add a color scalar value for the color processing stage to handle this case.
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
//					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATE);
					pDTDState->eColorMode = DTDState::EMode::kIterated;
				}
			}
		} else // In Tribes, we're always left with the other factor being ONE, so no blending.
		{ // We're simply deciding the draw modes without any form of blending.
				switch( color_other )
				{
				case GR_COMBINE_OTHER_CONSTANT: // Flat shaded mode.
					pDTDState->bUseTextures = false;
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
					pDTDState->eColorMode = DTDState::EMode::kConstant;
					break;

				case GR_COMBINE_OTHER_ITERATED: // Smooth shaded mode.
					pDTDState->bUseTextures = false;
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
					pDTDState->eColorMode = DTDState::EMode::kIterated;
					break;


				case GR_COMBINE_OTHER_TEXTURE: // Texture mapping mode.
					pDTDState->bUseTextures = true;
//					pDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);
//					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
					pDTDState->eColorMode = DTDState::EMode::kConstant;
					pDTDState->dwConstantColor |= 0x55555555;
					pDTDState->dwConstantColorNoAlpha |= 0x555555;
					break;
				}
		}

		// Always enabled the alpha blending.
		if(pDTDState->bUseBlending)
		{
			// Prepare the alpha processing stages.
			if( alpha_func == GR_COMBINE_FUNCTION_BLEND_OTHER )
			{ // Usual case, but we generally only have two modes in here.
				if( alpha_factor == GR_COMBINE_FACTOR_ONE )
				{ // We're working off of our Other only. (alpha = 1 * B)
					switch( alpha_other )
					{
					case GR_COMBINE_OTHER_CONSTANT: // Flat shaded mode.
						pDTDState->bUseTextures = true;
						pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
						pDTDState->eAlphaMode = DTDState::EMode::kConstant;
						break;

					case GR_COMBINE_OTHER_ITERATED: // Smooth shaded mode.
						pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
						pDTDState->eAlphaMode = DTDState::EMode::kIterated;
						break;


					case GR_COMBINE_OTHER_TEXTURE: // Texture mapping mode.
						pDTDState->bUseTextures = true;
						pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
						pDTDState->eAlphaMode = DTDState::EMode::kConstant;
						break;
					}
				} else
				{ // We're always using a factor of Local which is always constant, and the other is a Texture.  (alpha = A * B)
					pDTDState->bUseTextures = true;
					pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATEALPHA);
					pDTDState->eAlphaMode = DTDState::EMode::kConstant;
				}
			} else
			{ // The other mode is always Local and always constant. ( alpha = A )
				pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
				pDTDState->eAlphaMode = DTDState::EMode::kConstant;
			}
		} else
		{
			pDTDState->eAlphaMode = DTDState::EMode::kConstant;
		}

		// Override to not draw pixels (Max transparency)
//		if( pGlideState->mask_rgb == FXFALSE )
		{
//			pDTDState->bUseTextures = false;
			pDTDState->Framework.GetD3DDevice()->SetRenderState( D3DRENDERSTATE_SRCBLEND, D3DBLEND_SRCALPHA );
			pDTDState->Framework.GetD3DDevice()->SetRenderState( D3DRENDERSTATE_DESTBLEND, D3DBLEND_INVSRCALPHA );
		}
//		pDTDState->bUseTextures = true;
//		pDevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
//		pDevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);

		// Select the color processor.
		switch( pDTDState->eColorMode )
		{
		case DTDState::EMode::kIterated:
			if( pDTDState->eAlphaMode == DTDState::EMode::kIterated )
				VP_Color = VP_Color_II;
			else
				VP_Color = VP_Color_IC;
			break;

		case DTDState::EMode::kConstant:
			if( pDTDState->eAlphaMode == DTDState::EMode::kIterated )
				VP_Color = VP_Color_CI;
			else
				VP_Color = VP_Color_CC;
			break;

		case DTDState::EMode::kNone:
			VP_Color = VP_Null;
			break;
		}

		// Whenever we process our combine units, we can open up the possibility of using a texture, or turning texturing off.
		DirtyFlags |=  kTexture;
	}

	// Perform the actual texture setup opperations here.
	if( DirtyFlags & kTexture )
	{
		if( pDTDState->bUseTextures && pDTDState->pCurrentTexture)
		{
			VTexture *pTex = pDTDState->pCurrentTexture;

			// Select the texture, provided it still exists.
			if( pTex->pSurface && pTex->pTexture)
			{
				// If the texture is palettized then see if the texture palette needs updating.
				if( pDTDState->bPalDirty && pTex->bPal)
				{
					pTex->UpdatePal();
					pDTDState->bPalDirty = false;
				}

				// If we're in a chromakey mode, let's set it up with the texture surface.
				if( chroma_mode == GR_CHROMAKEY_ENABLE )
				{
					DDCOLORKEY ddck;
					ddck.dwColorSpaceLowValue =	ddck.dwColorSpaceHighValue = chroma_value;
					pDTDState->Framework.GetD3DDevice()->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE, TRUE);

					// TBI - note, does crash here if we don't have a valid pSurface. (See Above)
					pTex->pSurface->SetColorKey(DDCKEY_SRCBLT, &ddck);
				} else
				{
					pDTDState->Framework.GetD3DDevice()->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE, FALSE);
				}


				// Correct for the aspect ratio.  (Apparently 3Dfx's hardware really
				//   depends upon the long side's scale factor to represent 0-256 versus
				//   the short side's scale to 0-256)
				int nAspectW, nAspectH;
				ConvAspectToInt(nAspectW, nAspectH, pTex->m_Info.aspectRatio);

				pDTDState->fTexWidth = (float)nAspectH/256.f;
				pDTDState->fTexHeight = (float)nAspectW/256.f;

				// Finally set the texture we are to draw with.
				pDTDState->Framework.GetD3DDevice()->SetTexture(0, pTex->pTexture);
			} else
			{ // Invalid texture selection, so just set the device's texture to null.
				pDTDState->Framework.GetD3DDevice()->SetTexture(0, NULL );
			}
		} else
		{ // We don't use textures, so just set the device's texture to null.
			pDTDState->Framework.GetD3DDevice()->SetTexture(0, NULL );
		}
	}

	//-----------------------------------
	// Setup the various vertex processor stages.

	// Select the depth processor, and texture coordinate processor based on the depth mode.
	switch(	depth_mode )
	{
	case GR_DEPTHBUFFER_WBUFFER:
	case GR_DEPTHBUFFER_WBUFFER_COMPARE_TO_BIAS:
		VP_Depth = VP_Depth_WBuffer;
		VP_Texture = VP_Texture_W;
		break;

	case GR_DEPTHBUFFER_ZBUFFER_COMPARE_TO_BIAS:
	case GR_DEPTHBUFFER_ZBUFFER:
		VP_Depth = VP_Depth_ZBuffer;
		VP_Texture = VP_Texture_W;
		break;

	default:
		VP_Depth = VP_Null;
		VP_Texture = VP_Texture_No_W;
	}

	// We've selected the texture coordinate processor above, but if we're not using textures then override it.
	if( pDTDState->bUseTextures == false)
	{
		VP_Texture = VP_Texture_No_W;
	}

	DirtyFlags = 0;
}

// Glide state container class, used to represent the GrState pointer below.
// ** We have 312 bytes of freedom.  Use it wisely. That's 78 DWORDs... Or, 78 variables.
typedef struct _GlideStateContainer_tag{
	// Blend.
	GrAlphaBlendFnc_t	alpha_rgb_source;
	GrAlphaBlendFnc_t	alpha_rgb_dest;
	GrAlphaBlendFnc_t	alpha_alpha_source;
	GrAlphaBlendFnc_t	alpha_alpha_dest;

	// Alpha Combine.
	GrCombineFunction_t	alpha_func;
	GrCombineFactor_t	alpha_factor;
	GrCombineLocal_t	alpha_local;
	GrCombineOther_t	alpha_other;
	FxBool				alpha_invert;

	// Chromakey, in the alpha stage.
	GrCmpFnc_t	alpha_test_func;
	GrAlpha_t	alpha_test_value;

	// Chromakey, in the pixelFX stage.
	GrChromakeyMode_t	chroma_mode;
	GrColor_t			chroma_value;

	// Color combine.
	GrCombineFunction_t	color_func;
	GrCombineFactor_t	color_factor;
	GrCombineLocal_t	color_local;
	GrCombineOther_t	color_other;
	FxBool				color_invert;

	// Write masking enable/disable. (ColorMask, and DepthMask)
	FxBool mask_rgb, mask_alpha;

	// Constant color value.
	GrColor_t color_value;

	// Backface culling.
	GrCullMode_t cull_mode;

	// Depth info.
	FxI16				depth_bias;
	GrCmpFnc_t			depth_func;
	GrDepthBufferMode_t	depth_mode;
	FxBool				depth_mask;

	// Texture stuff not implemented yet... but pads nicely. <yep, I'm tired>
	FxU32 texaddress; // 32 bits
	GrTextureFilterMode_t tex_magfilter;
	GrTextureFilterMode_t tex_minfilter;
	GrMipMapMode_t mipmap_mode;
	FxBool mipmap_lodblend;
	float u;
	float v;
} GlideStateContainer;

// Store all the state info to the "state" structure.
void GlideState::GetState( GrState *state )
{
//	int gsc = sizeof( GlideState );
//	Assert( sizeof( GlideState ) < 312 );  // Size of structure exceeds padding space.

	// Simply copy our state info.
	memcpy( state, this, 312 );

//	GlideStateContainer* pState = (GlideStateContainer*)state;
//	int gsc = sizeof( GlideStateContainer );
/*
	// Alpha blending.
	pState->alpha_rgb_source	= alpha_rgb_source;
	pState->alpha_rgb_dest		= alpha_rgb_dest;
	pState->alpha_alpha_source	= alpha_alpha_source;
	pState->alpha_alpha_dest	= alpha_alpha_dest;

	// Alpha combine processing.
	pState->alpha_func		= alpha_func;
	pState->alpha_factor	= alpha_factor;
	pState->alpha_local		= alpha_local;
	pState->alpha_other		= alpha_other;
	pState->alpha_invert	= alpha_invert;

	// Chromakey, in the alpha stage.
	pState->alpha_test_func		= alpha_test_func;
	pState->alpha_test_value	= alpha_test_value;

	// Chromakey, in the pixelFX stage.
	chroma_mode		= chroma_mode;
	chroma_value	= chroma_value;

	// Write masking enable/disable. (ColorMask, and AlphaMask)
	pState->mask_rgb		= mask_rgb;
	pState->mask_alpha		= mask_alpha;

	pState->color_func		= color_func;
	pState->color_factor	= color_factor;
	pState->color_local		= color_local;
	pState->color_other		= color_other;
	pState->color_invert	= color_invert;

	pState->depth_bias		= depth_bias;
	pState->depth_mode		= depth_mode;
	pState->depth_func		= depth_func;
	pState->depth_mask		= depth_mask;

	pState->cull_mode		= cull_mode;
*/

	// TBI - texture management.
	/*
	FxU32 texaddress; // 32 bits
	GrTextureFilterMode_t tex_magfilter;
	GrTextureFilterMode_t tex_minfilter;
	GrMipMapMode_t mipmap_mode;
	FxBool mipmap_lodblend;
	float u;
	float v;
	*/
}

// Define all the export function prototypes here,
// because this is the only place we'll ever need them.
EXPORT void STD grAlphaBlendFunction(	GrAlphaBlendFnc_t rgb_sf,
										GrAlphaBlendFnc_t rgb_df,
										GrAlphaBlendFnc_t alpha_sf,
										GrAlphaBlendFnc_t alpha_df );

EXPORT void STD grAlphaCombine(	GrCombineFunction_t func,
							GrCombineFactor_t factor,
							GrCombineLocal_t local,
							GrCombineOther_t other,
							FxBool invert );

EXPORT void STD grAlphaTestFunction( GrCmpFnc_t function );
EXPORT void STD grAlphaTestReferenceValue( GrAlpha_t value );


EXPORT void STD grColorCombine(	GrCombineFunction_t func,
							GrCombineFactor_t factor,
							GrCombineLocal_t local,
							GrCombineOther_t other,
							FxBool invert );

EXPORT void STD grColorMask( FxBool rgb, FxBool alpha );
EXPORT void STD grConstantColorValue( GrColor_t color );

EXPORT void STD grChromakeyMode( GrChromakeyMode_t mode );
EXPORT void STD grChromakeyValue( GrColor_t value );

EXPORT void STD grDepthBiasLevel( FxI16 level );
EXPORT void STD grDepthBufferFunction( GrCmpFnc_t func );
EXPORT void STD grDepthBufferMode( GrDepthBufferMode_t mode );
EXPORT void STD grDepthMask( FxBool enable );

EXPORT void STD grCullMode( GrCullMode_t mode );

EXPORT void STD grRenderBuffer( GrBuffer_t buffer );

// Re-enable all the active state info.
void GlideState::SetState( const GrState *state )
{
	// Make sure we're not in a scene drawing state.
	pDTDState->EndScene();

	// Restore the scene state through the container struct.
	GlideState* pState = (GlideState*)state;

	// Call the functions to reset the state variables.
	grAlphaBlendFunction( pState->alpha_rgb_source, pState->alpha_rgb_dest,
						  pState->alpha_alpha_source, pState->alpha_alpha_dest );

	grAlphaTestFunction( pState->alpha_test_func );
	grAlphaTestReferenceValue( pState->alpha_test_value );

	grAlphaCombine( pState->alpha_func, pState->alpha_factor, pState->alpha_local, pState->alpha_other, pState->alpha_invert );
	grColorCombine( pState->color_func, pState->color_factor, pState->color_local, pState->color_other, pState->color_invert );

	grColorMask( pState->mask_rgb, pState->mask_alpha );

	grConstantColorValue( pState->color_value );

	grChromakeyMode( pState->chroma_mode );
	grChromakeyValue( pState->chroma_value );


	grDepthBiasLevel( pState->depth_bias );
	grDepthBufferMode( pState->depth_mode );
	grDepthBufferFunction( pState->depth_func );
	grDepthMask( pState->depth_mask );

	grCullMode( pState->cull_mode );

	grRenderBuffer( pState->render_buffer );
}
