//-------------------------------------------------------------------
//	File:		VertexProcessor.cpp
//	Created:	23/02/99 2:52AM
//	Author:		Aaron Hilton
//	Comments:	Process a Glide vertex into Direct3D format in the most optimal form.
//  Thanks to Scott Cutler for the wonderful vertex processing idea!
//-------------------------------------------------------------------
#include "stdafx.h"
#include "VertexProcessor.h"

// Currently selected vertex processing functions.
void (*VP_Snap)(); // Pointer to the vertex snapping funciton.
void (*VP_Depth)(); // Depth buffer vertex processing.
void (*VP_Color)(); // Pointer to the color processing function.
void (*VP_Texture)(); // Pointer to the texture coordinate calculation function.


//-------------------------------------------------------------------
// Vertex buffer, keep the processed verticies.
D3DTLVERTEX vertexBuffer[MAX_VERTS];
// General variables for remapped functions.
static const GrVertex *VP_Src;
static D3DTLVERTEX *VP_Dest;

const float vertex_snap = float(3L << 18);

//-------------------------------------------------------------------
// Generic NULL function pointer.
void VP_Null()
{
}

//-------------------------------------------------------------------
// Perform vertex snapping.
void VP_P_Snap()
{
	VP_Dest->sx = VP_Src->x - vertex_snap;
	VP_Dest->sy = VP_Src->y - vertex_snap;
}

// Just copy the vertex coordinate.
void VP_P_NoSnap()
{
	VP_Dest->sx = VP_Src->x;
	VP_Dest->sy = VP_Src->y;
}

// Perform vertex snapping with origin in lower left.
void VP_P_Snap_Flip()
{
	VP_Dest->sx = VP_Src->x - vertex_snap;
	VP_Dest->sy = pDTDState->fViewportHeight - (VP_Src->y - vertex_snap);
}

// Just copy the vertex coordinates with origin in lower left (flip Y).
void VP_P_NoSnap_Flip()
{
	VP_Dest->sx = VP_Src->x;
	VP_Dest->sy = pDTDState->fViewportHeight - VP_Src->y;
}

//-------------------------------------------------------------------
// Depth buffer vertex processing.

// W-Buffer conversion.  (Scott Cutler's method)
void VP_Depth_WBuffer()
{
	float w = 1.0f / VP_Src->oow;
	VP_Dest->sz = float(((*(DWORD *)&w >> 11) & 0xFFFFF) - (127 << 12)) / 65536.0f;
	VP_Dest->sz += ((float)pDTDState->fZBias);
}

// Regular Z-Buffer conversion.
void VP_Depth_ZBuffer()
{
	VP_Dest->sz = VP_Src->ooz*65536.0f;
	VP_Dest->sz += ((float)pDTDState->fZBias);
}

//-------------------------------------------------------------------
// Color processing.

// Color Iterated, and Alpha Iterated.
void VP_Color_II()
{
	VP_Dest->color = RGBA_MAKE(int(VP_Src->r), int(VP_Src->g), int(VP_Src->b), int(VP_Src->a));
}

// Color Iterated, and Alpha Constant.
void VP_Color_IC()
{
	VP_Dest->color = RGBA_MAKE(int(VP_Src->r), int(VP_Src->g), int(VP_Src->b), int(pDTDState->dwConstantAlpha));
}

// Color Constant, and Alpha Iterated.
void VP_Color_CI()
{
	VP_Dest->color = pDTDState->dwConstantColorNoAlpha | ((DWORD)VP_Src->a<<24);
}

// Color Constant, and Alpha Constant.
void VP_Color_CC()
{
	VP_Dest->color = pDTDState->dwConstantColor;
}

//-------------------------------------------------------------------
// Texture coordinate processing.

// We have embeded w information in each vertex.
void VP_Texture_W()
{
	float w = 1.0f / VP_Src->oow;
	VP_Dest->tu = (VP_Src->tmuvtx[0].sow * w) * pDTDState->fTexWidth;
	VP_Dest->tv = (VP_Src->tmuvtx[0].tow * w) * pDTDState->fTexHeight;
	VP_Dest->rhw = VP_Src->oow;
}

// No W information to use, just default to 1.0
void VP_Texture_No_W()
{
	VP_Dest->tu = VP_Src->tmuvtx[0].sow * pDTDState->fTexWidth;
	VP_Dest->tv = VP_Src->tmuvtx[0].tow * pDTDState->fTexHeight;
	VP_Dest->rhw = 1.f;
}

//-------------------------------------------------------------------
// Perform all the stages of vertex processing.
void VertexProcessor( const GrVertex *vert, int nOffset )
{
	VP_Src = vert;
	VP_Dest=&vertexBuffer[nOffset];

	VP_Snap();
	VP_Depth();

	VP_Dest->specular = 0x88888888;
	VP_Color();

	VP_Texture();
}

//-------------------------------------------------------------------
//	History:
//	$Log:$
//-------------------------------------------------------------------
